#include <cassert>
#include <string>
#include <sstream>
#include <iostream>
#include "Move.h"
#include "Board.h"
#include "Position.h"
#include "Eval.h"
#include "Zobrist.h"


using namespace std;

Position::Position() {
}

Position::Position(Position& position)
{
	board = position.board;
	gameStateStack = position.gameStateStack;
}

Position::Position(string strPos)
{
	SetFromFEN(strPos);
}

// get zobrist key
Key Position::GetKey()
{
  return gameState.zobrist;
}     

void Position::SetFromFEN(string fenString)
{
	stringstream fen(fenString); // <-- use this sstream instead !!!
	fen >> std::noskipws;
  
	board.Clear();
	gameStateStack.clear();
	Square square = SQ_A8;

	char carBoard = '#';
	while(carBoard != ' ')
	{
		fen >> carBoard ;
		switch (carBoard)
		{
		case 'K':
			board.SetPiece(W_KING, square);
			square++;
			break;
		case 'Q':
			board.SetPiece(W_QUEEN, square);
			square++;
			break;
		case 'R':
			board.SetPiece(W_ROOK, square);
			square++;
			break;
		case 'N':
			board.SetPiece(W_KNIGHT, square);
			square++;
			break;
		case 'B':
			board.SetPiece(W_BISHOP, square);
			square++;;
			break;
		case 'P':
			board.SetPiece(W_PAWN, square);
			square++;
			break;
		case 'k':
			board.SetPiece(B_KING, square);
			square++;
			break;
		case 'q':
			board.SetPiece(B_QUEEN, square);
			square++;
			break;
		case 'r':
			board.SetPiece(B_ROOK, square);
			square++;
			break;
		case 'n':
			board.SetPiece(B_KNIGHT, square);
			square++;
			break;
		case 'b':
			board.SetPiece(B_BISHOP, square);
			square++;
			break;
		case 'p':
			board.SetPiece(B_PAWN, square);
			square++;
			break;
		case '/':
			square -= (Square)16;
			break;
		case '1':
			square += (Square)1;
			break;
		case '2':
			square += (Square)2;
			break;
		case '3':
			square += (Square)3;
			break;
		case '4':
			square += (Square)4;
			break;
		case '5':
			square += (Square)5;
			break;
		case '6':
			square += (Square)6;
			break;
		case '7':
			square += (Square)7;
			break;
		case '8':
			square += (Square)8;
			break;
		default:
			//goto endBoard;
			break;
		}
	}

	char ply = carBoard;
	fen >> ply;
	//ply = fenString[index++];
	switch (ply)
	{
	case 'w':
		gameState.turn = WHITE;
		break;
	case 'b':
		gameState.turn = BLACK;
		break;
	default:
		break;
	}
	fen >> ply;

	gameState.whiteCanCastleKingSide = false;
	gameState.whiteCanCastleQueenSide = false;
	gameState.blackCanCastleKingSide = false;
	gameState.blackCanCastleQueenSide = false;
	char carCastle = '#';
	while(carCastle != ' ')
	{
		fen >> carCastle;
		switch (carCastle)
		{
		case 'K':
			gameState.whiteCanCastleKingSide = true;
			break;
		case 'Q':
			gameState.whiteCanCastleQueenSide = true;
			break;
		case 'k':
			gameState.blackCanCastleKingSide = true;
			break;
		case 'q':
			gameState.blackCanCastleQueenSide = true;
			break;
		default:
			break;
		}
	}

	char enPassant1, enPassant2;
	fen >> enPassant1;
	if (enPassant1 != '-')
	{
		fen >> enPassant2;
		gameState.lastEnPassant = make_square((File)(enPassant1- 'a'),(Rank)(enPassant2 - '1'));
	} else {
		gameState.lastEnPassant = SQ_NONE;
	}

	fen >> std::skipws;

	fen >> gameState.halfMoveClock;
	fen >> gameState.fullMoveNumber;
  
  // Generate zobrist key
  gameState.zobrist = book_key(*this);

}


// Perform the move, augment gameStack and change turn
void Position::Play(Move m)
{
	assert(from_sq(m) != to_sq(m));
	assert(m != MOVE_NULL);

	Piece piece = board.GetPiece(from_sq(m));
	Piece cPiece = board.GetPiece(to_sq(m));
  
	// push the previous gamestate to the stack 
	gameStateStack.push_back(gameState);

	gameState.playedPiece = piece;
	gameState.capturedPiece = cPiece;
  
	// Castleling - canceller
	// TODO: find a clean and quick way to update this status
	if (piece == W_KING)
	{
		gameState.whiteCanCastleKingSide = false;
		gameState.whiteCanCastleQueenSide = false;
	}
	else if (piece == B_KING)
	{
		gameState.blackCanCastleKingSide = false;
		gameState.blackCanCastleQueenSide = false;
	}
	else if (piece == W_ROOK && from_sq(m) == SQ_A1)
	{
		gameState.whiteCanCastleQueenSide = false;
	}
	else if (piece == W_ROOK && from_sq(m) == SQ_H1)
	{
		gameState.whiteCanCastleKingSide = false;
	}
	else if (piece == B_ROOK && from_sq(m) == SQ_A8)
	{
		gameState.blackCanCastleQueenSide = false;
	}
	else if (piece == B_ROOK && from_sq(m) == SQ_H8)
	{
		gameState.blackCanCastleKingSide = false;
	}

	// keeping en-passant last move
	if ((piece == W_PAWN && (rank_of(to_sq(m)) - rank_of(from_sq(m))) == 2)
		|| (piece == B_PAWN && (rank_of(to_sq(m)) - rank_of(from_sq(m))) == 2))
	{
		gameState.lastEnPassant = (from_sq(m)<32)?(from_sq(m)+DELTA_N):(from_sq(m)+DELTA_S);
	}
	else
	{
		gameState.lastEnPassant = SQ_NONE;
	}

	// standard move
	board.DoMove(m);

  int turn = gameState.turn;
  gameState.turn = (BLACK-gameState.turn);
  // 50 plies rule
  if (cPiece != NO_PIECE
			|| piece == W_PAWN
			|| piece == B_PAWN)
    gameState.halfMoveClock = 0;
  else
    gameState.halfMoveClock++;
  // full move clock (+1 only after black ply)
  if(turn == BLACK)
    gameState.fullMoveNumber++;
  
  // TODO: udpdate by delta instead of using book_key (rather costly)
  gameState.zobrist = book_key(*this);
}

void Position::Unplay(Move m)
{
  assert(m != MOVE_NULL);
	assert(this->gameStateStack.size() >= 1);
	
  // Unplay - base part
	board.UndoMove(m, gameState.playedPiece,gameState.capturedPiece);

	// recover previous gamestate
	gameState = gameStateStack.back();

	// Pop the state to return to the n-1 game state
	gameStateStack.pop_back();
	
	// Managed by gamestate
  //zobrist = book_key(*this);
}

void Position::SetPiece(Piece piece, Square square)
{
	// TODO : improve this code
	if (piece == NO_PIECE)
	{
		board.ClearSquare(square);
	}
	else
	{
		board.SetPiece(piece, square);
	}
}


void Position::SetStartPosition()
{
	this->SetFromFEN("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1");
}


string Position::GetFEN()
{
	stringstream result;
	int emptySquares = 0;
	for (Square i = SQ_A8; i >= SQ_A1; i += DELTA_S)
	{
		for (Square j = (Square)0; j < 8; j++)
		{

			Piece p = board.GetPiece(i + j);
			if (p == NO_PIECE)
			{
				emptySquares++;
			}
			else
			{
				if (emptySquares > 0)
				{
					result << emptySquares;
					emptySquares = 0;
				}
				result << " PNBRQK  pnbrqk  "[(int)p];

			}
		}
		if (emptySquares > 0)
		{
			result << emptySquares;
			emptySquares = 0;
		}
		if (i != 0)
			result << '/';
	}

	result << ' ';
	result << (GetTurn()==0 ? 'w' : 'b'); //WHITE==0
	result << ' ';
	
  if (gameState.whiteCanCastleKingSide)   result << 'K';
	if (gameState.whiteCanCastleQueenSide)  result << 'Q';
	if (gameState.blackCanCastleKingSide)   result << 'k';
	if (gameState.blackCanCastleQueenSide)  result << 'q';
  
	if (!gameState.whiteCanCastleKingSide
		&& !gameState.whiteCanCastleQueenSide
		&& !gameState.blackCanCastleKingSide
		&& !gameState.blackCanCastleQueenSide)
	result << '-';

	result << ' ';

	if (gameState.lastEnPassant != SQ_NONE)
    result << square_to_string(gameState.lastEnPassant);
	else
    result << '-';

	result << ' ';
	result << gameState.halfMoveClock;
	result << ' ';
	result << gameState.fullMoveNumber;

	return result.str();

}


// for test only !!!
bool Position::IsEqual(Position& pos)
{
	return board.IsEqual(pos.board);//		&& IsEqual(gameStateStack,pos.gameStateStack);
}


int Position::Evaluate()
{
	return Evaluate(GetTurn());
}

int Position::Evaluate(Color color)
{
	if(color == WHITE)
		return EvaluateBoard(board);
	return - EvaluateBoard(board);
}


int Position::GetPlayedMoveNumber()
{
	return (int)gameStateStack.size();
}


void Position::DebugDisplay()
{
	if(this->GetTurn() == WHITE)
		std::cout <<  "White to move" << std::endl;
	else
		std::cout <<  "Black to move" << std::endl;

	std::cout <<  "Castle rights: [";

	if(gameState.whiteCanCastleKingSide)
		std::cout <<  'K';
	if(gameState.whiteCanCastleQueenSide)
		std::cout <<  'Q';
	if(gameState.blackCanCastleKingSide)
		std::cout <<  'k';
	if(gameState.blackCanCastleQueenSide)
		std::cout <<  'q';
	std::cout <<  ']' << std::endl;

	board.DebugDisplay();
}

bool Position::IsCheck()
{
	return board.IsCheck(this->GetTurn());
}

bool Position::IsCheck(Color side)
{
	return board.IsCheck(side);
}

bool Position::IsCheck(Color side, Square square)
{
	return board.IsCheckOnSquare(square, BLACK-side);
}

